#!/usr/bin/perl
# fedora-extras version

use strict;

# apparently you aren't supposed to leave warnings
# on when you distribute programs...
# use warnings;
$|++;

use Cwd;
use locale;
use POSIX qw/locale_h strftime/;
use File::Basename;
use File::Find::Rule;
use File::Path qw/mkpath/;
use Date::Calc qw/Delta_Days/;

use Gtk2;
use Gtk2::SimpleList;
use Gtk2::SimpleMenu;
use Glib qw/TRUE FALSE/;
Gtk2->init;

use constant COLUMN_FILE   => 0;
use constant COLUMN_TYPE   => 1;
use constant COLUMN_SIZE   => 2;
use constant COLUMN_STATUS => 3;
use constant NUM_COLUMNS   => 4;

my $VERSION   = '2.13';
my $virus_log = '';
my ( $key, $value );
my $ren = "None";
my ( $detect_broken, $save_log, $hidden, $showall ) = (0) x 4;
my @virus;
my $count       = 0; # keeps track of the viruses (e.g., $virus[$count]{full})
my $num_scanned = 0; # counts number of files scanned
my $num_so_far  = 0; # number of viruses
my %found;
my $start_time;
my ( $q_state, $l_state );
my ( @files,   @quoted );
my $directory = $ENV{HOME} || getcwd();
my $c_dir     = "$directory/.clamtk";
my $v_dir     = "$c_dir/viruses";
my $l_dir     = "$c_dir/history";
my ( $a_tooltip, $authenticate );
my %dirs_scanned;
my $scan_pid = '';

# maintenance subroutine variables below
my $new_win;
my ( $new_slist, $new_hlist );
my @q_files = ();
my $q_label;
my @h_files = ();
my $h_label;

BEGIN {
    my $encoding = setlocale(LC_CTYPE);
    $ENV{LC_ALL} = $encoding;
}

if ( $> == 0 ) {
    $authenticate = 'gtk-yes';
    $a_tooltip    = "Check for signature updates";
}
else {
    $authenticate = 'gtk-no';
    $a_tooltip    = "You must be root to install updates";
}

my $command;

my $FRESHPATH =
      ( -e '/usr/bin/freshclam' ) ? '/usr/bin/freshclam'
    : ( -e '/usr/local/bin/freshclam' ) ? '/usr/local/bin/freshclam'
    : die "freshclam not found!\n";
my $SIGPATH =
      ( -e '/usr/bin/sigtool' ) ? '/usr/bin/sigtool'
    : ( -e '/usr/local/bin/sigtool' ) ? '/usr/local/bin/sigtool'
    : die "sigtool not found!\n";
my $CLAMPATH =
      ( -e '/usr/bin/clamscan' ) ? '/usr/bin/clamscan'
    : ( -e '/usr/local/bin/clamscan' ) ? '/usr/local/bin/clamscan'
    : die "clamscan not found!\n";
$command .= $CLAMPATH;
my $RARPATH =
      ( -e '/usr/bin/unrar' ) ? '/usr/bin/unrar'
    : ( -e '/usr/local/bin/unrar' ) ? '/usr/local/bin/unrar'
    : '';

$command .= " --unrar=$RARPATH" if ($RARPATH);
my $ZIPPATH =
      ( -e '/usr/bin/unzip' ) ? '/usr/bin/unzip'
    : ( -e '/usr/local/bin/unzip' ) ? '/usr/local/bin/unzip'
    : '';

$command .= " --unzip=$ZIPPATH" if ($ZIPPATH);
my $FILE =
      ( -e '/usr/bin/file' ) ? '/usr/bin/file'
    : ( -e '/usr/local/bin/file' ) ? '/usr/local/bin/file'
    : die "\"file\" command not found!\n";

# virus definitions path
my $DEFPATH =
      ( -e '/var/lib/clamav/daily.cvd' ) ? '/var/lib/clamav/daily.cvd'
    : ( -e '/var/clamav/daily.cvd' ) ? '/var/clamav/daily.cvd'
    : '';

$command .= " --no-summary ";

if ( !-d $v_dir ) {
    eval { mkpath( $v_dir, 0, 0777 ); };
    if ($@) {
        $q_state = "disabled";
    }
    else {
        $q_state = "normal";
    }
}
else {
    $q_state = "normal";
}

if ( !-d $l_dir ) {
    eval { mkpath( $l_dir, 0, 0777 ); };
    if ($@) {
        $l_state = "disabled";
    }
    else {
        $l_state = "normal";
    }
}
else {
    $l_state = "normal";
}

my $window = Gtk2::Window->new();
$window->signal_connect( destroy => sub { Gtk2->main_quit; } );
$window->set_default_size( 630, 325 );
$window->set_title("ClamTk Virus Scanner");
$window->set_border_width(0);

my $main_vbox = Gtk2::VBox->new( FALSE, 0 );
$window->add($main_vbox);
$main_vbox->show;

my $menu_tree = create_menu();

my $menu = Gtk2::SimpleMenu->new(
    menu_tree        => $menu_tree,
    default_callback => \&return,
    user_data        => 'user data',
);

$window->add_accel_group( $menu->{accel_group} );
my $menubar = $menu->{widget};
$main_vbox->pack_start( $menubar, FALSE, FALSE, 1 );

# These are the GUI's for scanning and exiting
my $toolbar = Gtk2::Toolbar->new;
$toolbar->set_style('icons');
my $tt = Gtk2::Tooltips->new();

# We can set the size of the toolbar stuff here, but
# for now, I like the default
# $toolbar->set_icon_size('menu');
my $scan_file = Gtk2::ToolButton->new_from_stock('gtk-new');
$scan_file->signal_connect( 'clicked' => sub { getfile('file') } );
$scan_file->set_tooltip( $tt, "Scan a file", "" );
$toolbar->insert( $scan_file, -1 );
my $scan_dir = Gtk2::ToolButton->new_from_stock('gtk-directory');
$scan_dir->signal_connect( 'clicked' => sub { getfile('dir') } );
$scan_dir->set_tooltip( $tt, "Scan a directory", "" );
$toolbar->insert( $scan_dir, -1 );
my $scan_home = Gtk2::ToolButton->new_from_stock('gtk-home');
$scan_home->signal_connect( 'clicked' => sub { getfile('home') } );
$scan_home->set_tooltip( $tt, "Scan your home directory", "" );
$toolbar->insert( $scan_home, -1 );
$toolbar->insert( Gtk2::SeparatorToolItem->new, -1 );
my $scan_clear = Gtk2::ToolButton->new_from_stock('gtk-clear');
$scan_clear->signal_connect( 'clicked' => sub { clear_output(); } );
$scan_clear->set_tooltip( $tt, "Clear the display", "" );
$toolbar->insert( $scan_clear, -1 );
my $scan_stop = Gtk2::ToolButton->new_from_stock('gtk-stop');
$scan_stop->signal_connect(
    'clicked' => sub { kill 15, $scan_pid if ($scan_pid) } );
$scan_stop->set_tooltip( $tt, "Stop scanning now", "" );
$toolbar->insert( $scan_stop, -1 );
my $scan_exit = Gtk2::ToolButton->new_from_stock('gtk-quit');
$scan_exit->signal_connect( 'clicked' => sub { Gtk2->main_quit } );
$scan_exit->set_tooltip( $tt, "Quit", "" );
$toolbar->insert( $scan_exit, -1 );
$main_vbox->pack_start( $toolbar, FALSE, FALSE, 1 );

# This is the top label where scanning messages are displayed
my $top_label = Gtk2::Label->new();
$main_vbox->pack_start( $top_label, FALSE, FALSE, 3 );
$top_label->set_justify('center');

# This scrolled window holds the slist
my $scrolled_win = Gtk2::ScrolledWindow->new;
$scrolled_win->set_shadow_type('etched_in');
$scrolled_win->set_policy( 'automatic', 'automatic' );
$main_vbox->pack_start( $scrolled_win, TRUE, TRUE, 0 );

my $slist = create_list();
$scrolled_win->add($slist);
$scrolled_win->grab_focus();
$slist->get_selection->set_mode('single');
$slist->set_rules_hint(TRUE);

# can't be reorderable - messes up the \&row_clicked function
$slist->set_reorderable(FALSE);
map { $_->set_sizing('autosize') } $slist->get_columns;

# this anonymous sub handles the row_clicked feature
$slist->get_selection->signal_connect(
    changed => sub {
        my @sel   = $slist->get_selected_indices;
        my $deref = $sel[0];
        defined $deref or return;
        $top_label->set_markup(
            "<b>File:</b> $virus[$deref]{full}    <b>Status:</b> $virus[$deref]{status}"
        );
    }
);

# below: the right-click functionality. also uses 'sub confirm'.
$slist->signal_connect(
    button_press_event => sub {
        my ( $widget, $event ) = @_;
        return FALSE unless $event->button == 3;
        my @sel   = $slist->get_selected_indices;
        my $deref = $sel[0];
        defined $deref or return;

        my $menu     = Gtk2::Menu->new();
        my $quar_pop = Gtk2::MenuItem->new('Quarantine this file');
        $quar_pop->signal_connect( activate => sub { confirm( $deref, "q" ) }
        );
        $quar_pop->show();
        $menu->append($quar_pop);

        my $delete_pop = Gtk2::ImageMenuItem->new('Delete this item');
        my $del_image = Gtk2::Image->new_from_stock( 'gtk-delete', 'menu' );
        $delete_pop->set_image($del_image);
        $delete_pop->signal_connect(
            activate => sub { confirm( $deref, "d" ) } );
        $delete_pop->show();
        $menu->append($delete_pop);

        my $cancel_pop
            = Gtk2::ImageMenuItem->new_from_stock( 'gtk-cancel', undef );
        $cancel_pop->signal_connect( activate => sub { return; } );
        $cancel_pop->show();
        $menu->append($cancel_pop);
        $menu->popup( undef, undef, undef, undef, $event->button,
            $event->time );
    }
);

# bottom_box keeps track of # scanned, # of viruses, and time
my $bottom_box = Gtk2::HBox->new( FALSE, 0 );
$main_vbox->pack_start( $bottom_box, FALSE, FALSE, 2 );

my $left_status = Gtk2::Label->new("Files Scanned: ");
$left_status->set_alignment( 0.0, 0.5 );
$bottom_box->pack_start( $left_status, TRUE, TRUE, 4 );

my $mid_status = Gtk2::Label->new("Viruses Found: ");
$mid_status->set_alignment( 0.0, 0.5 );
$bottom_box->pack_start( $mid_status, TRUE, TRUE, 0 );

my $right_status = Gtk2::Label->new("Ready");
$right_status->set_alignment( 0.0, 0.5 );
$bottom_box->pack_start( $right_status, TRUE, TRUE, 0 );

$window->show_all();

# This is to combine any and all startup checks
startup_prefs();

Gtk2->main;

sub about {
    my $message = "ClamTk Virus Scanner\nVersion $VERSION\n"
        . "http://freshmeat.net/projects/clamtk/\nhttp://clamtk.sourceforge.net\n\n"
        . "Copyright (c) 2004-2006 Dave M\n(dave.nerd AT gmail DOT com)\n\n"
        . "This program is free software.\nYou may modify it and/or redistribute it\n"
        . "under the same terms as Perl itself.\n\nPlease read the DISCLAIMER, LICENSE,\n"
        . "and README for more information.\n\nClam AntiVirus is available at http://www.clamav.net .\t\n";
    my $dialog = Gtk2::MessageDialog->new_with_markup( $window,
        [qw(modal destroy-with-parent)],
        'info', "ok", $message );

    if ( "ok" eq $dialog->run ) {
        $dialog->destroy;
        return;
    }

}

sub create_list {
    my $list = Gtk2::SimpleList->new(
        'File'   => 'markup',
        'Type'   => 'markup',
        'Size'   => 'markup',
        'Status' => 'markup',
    );

    return $list;
}

sub getfile {
    my $option = shift;

    # $option will be either "home", "file", "dir", or "recur"
    Gtk2->main_iteration while ( Gtk2->events_pending );
    clear_output();
    chdir($directory) or chdir("/tmp");

    my ( $filename, $dir, $dialog );
    if ( $option eq "home" ) {
        @files
            = File::Find::Rule->file()->readable->maxdepth(1)->in($directory);
    }
    elsif ( $option eq "file" ) {
        $dialog = Gtk2::FileChooserDialog->new(
            'Select a File', undef, 'open',
            'gtk-cancel' => 'cancel',
            'gtk-ok'     => 'ok',
        );

        if ( "ok" eq $dialog->run ) {
            $filename = $dialog->get_filename;
            $dialog->destroy;
            push( @files, $filename );
        }
        else {
            $dialog->destroy;
            return;
        }
    }
    elsif ( $option eq "dir" ) {
        $dialog = Gtk2::FileChooserDialog->new(
            'Select a Directory (directory scan)', undef, 'select-folder',
            'gtk-cancel' => 'cancel',
            'gtk-ok'     => 'ok',
        );

        if ( "ok" eq $dialog->run ) {
            $dir = $dialog->get_filename;
            $dialog->destroy;
            $dir ||= $directory;
            @files
                = File::Find::Rule->file()->readable->maxdepth(1)->in($dir);
        }
        else {
            $dialog->destroy;
            return;
        }
    }
    elsif ( $option eq "recur" ) {
        $dialog = Gtk2::FileChooserDialog->new(
            'Select a Directory (recursive scan)', undef, 'select-folder',
            'gtk-cancel' => 'cancel',
            'gtk-ok'     => 'ok'
        );

        if ( "ok" eq $dialog->run ) {
            $dir = $dialog->get_filename;
            $dialog->destroy;
            $dir ||= $directory;
            @files = File::Find::Rule->file()->readable->in($dir);
        }
        else {
            $dialog->destroy;
            return;
        }
    }
    else {
        $dialog->destroy;
        return;
    }

    # start the timer - replaces the "Ready"
    $start_time = time;
    $right_status->set_text("Elapsed Time: ");

    if ( $hidden == 0 ) {
        @files = grep { basename($_) !~ /^\./ } @files;
    }
    my @new;
    my @large;
    foreach my $foo (@files) {
        if ( -s $foo >= 15_000_000 ) {
            push( @large, $foo );
        }
        else {
            push( @new, $foo );
        }
    }
    if (@large) {
        foreach my $too_big (@large) {
            $top_label->set_text("Scanning $too_big...");
            timer();
            $num_scanned++;
            $virus[$count]{full}   = $too_big;
            $virus[$count]{base}   = basename($too_big);
            $virus[$count]{status} = "Not scanned (size)";
            $virus[$count]{type}   = type($too_big);
            $virus[$count]{size}   = size($too_big);
            display();
            next;
        }
    }
    if ( scalar(@new) ) {
        scan(@new);
    }
    clean_up();
}

sub scan {
    my @get = @_;
    @quoted = map { quotemeta($_) } @get;
    timer();

    # this goes so fast, the stop button is pointless unless
    # there are $lots of files to scan
    defined( my $pid = open( SCAN, "$command @quoted |" ) )
        or die "couldn't fork: $!\n";
    $scan_pid = $pid;    # this is for the 'stop button'
    while (<SCAN>) {
        Gtk2->main_iteration while Gtk2->events_pending;
        my ( $file, $status ) = split /:/;

        chomp($file)   if ( defined $file );
        chomp($status) if ( defined $status );
        next unless ( -e $file && $status );
        next if ( $status =~ /module failure/ );

        $dirs_scanned{ dirname($file) } = 1
            unless ( dirname($file) =~ /\/tmp\/clamav/
            || dirname($file) eq "." );

        $virus[$count]{base} = basename($file);
        if ( length( $virus[$count]{base} ) > 17 ) {
            $virus[$count]{base} = substr( $virus[$count]{base}, 0, 17 );
        }
        $top_label->set_text("Scanning $virus[$count]{base}...");
        $status =~ s/\s+FOUND$//;
        my $meta = quotemeta($file);

        $virus[$count]{type} = type($file);
        $virus[$count]{full} = $file;
        $virus[$count]{size} = size($file);

        # ignore files in archives - we just want the end-result.
        # we still allow it to scan; clamav will show the result
        # this method doesn't count it, though...
        next if ( $virus[$count]{full} =~ /\/tmp\/clamav/ );
        $num_scanned++;

        $virus[$count]{status} = $status;

        $left_status->set_text("Files Scanned: $num_scanned");
        timer();

        # clean_words mean no viruses... haven't seen any others than this
        my $clean_words = join( '|',
            "OK",                 "Zip module failure",
            "RAR module failure", "Encrypted.RAR",
            "Encrypted.Zip",      "Empty file" );

        if ( $status !~ /$clean_words/ ) {    # a virus
            $found{ $virus[$count]{full} } = $virus[$count]{status};
            if ( $ren eq "Quarantine" ) {
                if ( dirname( $virus[$count]{full} ) !~ /\/tmp\/clamav/ ) {
                    move_to_quarantine($count);
                    my $current_status = $virus[$count]{status};
                    $virus[$count]{status} = "$current_status (Quarantined)";
                }
            }
            elsif ( $ren eq "Delete" ) {

          # Can't use the "--remove" option with clamscan; the way this scans,
          # the file is already deleted before it's (sub) displayed...
                if ( dirname( $virus[$count]{full} ) !~ /\/tmp\/clamav/ ) {
                    unlink( $virus[$count]{full} );
                    my $current_status = $virus[$count]{status};
                    $virus[$count]{status} = "$current_status (Deleted)";
                }
            }
        }

        $num_so_far = keys %found;
        $left_status->set_text("Files Scanned: $num_scanned");
        if ( $num_so_far > 0 ) {
            $mid_status->set_markup("<b>Viruses Found: $num_so_far</b>");
        }
        else {
            $mid_status->set_text("Viruses Found: $num_so_far");
        }
        display();
    }
    close(SCAN);    # or warn "Unable to close scanner! $!\n";
}

sub display {
    timer();
    use encoding 'utf8';
    $virus[$count]{type}   =~ s/\s+$//;
    $virus[$count]{status} =~ s/\s+$//;
    $virus[$count]{status} =~ s/^\s//;
    if ((      $virus[$count]{status} ne "OK"
            && $virus[$count]{status} ne "Empty file"
        )
        || $showall
        )
    {
        push @{ $slist->{data} },
            [
            $virus[$count]{base}, $virus[$count]{type},
            $virus[$count]{size}, $virus[$count]{status}
            ];
        if ( $virus[$count]{status} =~ /(\(Quarantined\)|\(Deleted\))/ ) {
            slist_delete($count);
        }
        $count++;
    }
    timer();
}

sub timer {
    Gtk2->main_iteration while ( Gtk2->events_pending );
    my $now     = time;
    my $seconds = $now - $start_time;
    my $s       = sprintf "%02d", ( $seconds % 60 );
    my $m       = sprintf "%02d", ( $seconds - $s ) / 60;
    $right_status->set_text("Elapsed time: $m:$s");
    $window->queue_draw;
}

sub clean_up {
    $count ||= 0;

    my $db_total = num_of_sigs();
    if ($save_log) {
        my ( $mon, $day, $year ) = split / /,
            strftime( '%b %d %Y', localtime );
        $virus_log = "$mon-$day-$year" . ".log";

        # sort the directories scanned for display
        my @sorted = sort { length $a <=> length $b } keys %dirs_scanned;
        if ( open REPORT, ">>$l_dir/$virus_log" ) {
            print REPORT "\nClamTk, v$VERSION\n", scalar localtime, "\n";
            print REPORT "Current ClamAV Signatures: $db_total\n";
            print REPORT "Directories Scanned:\n";
            for my $list (@sorted) {
                print REPORT "$list\n";
            }
            print REPORT
                "\nFound $num_so_far possible viruses ($num_scanned files scanned).\n\n";
        }
        else {
            $top_label->set_text(
                "Could not write to logfile. Check permissions.");
            $save_log = 0;
        }

    }
    $db_total =~ s/(\w+)\s+$/$1/;
    $top_label->set_text("Scanning complete ($db_total signatures)");
    $left_status->set_text("Files Scanned: $num_scanned");
    if ( $num_so_far == 0 ) {
        $mid_status->set_text("Viruses Found: $num_so_far");
    }
    $right_status->set_text("Ready");
    $window->queue_draw;

    if ( $num_so_far == 0 ) {
        print REPORT "No viruses found.\n" if ($save_log);
    }
    else {
        if ($save_log) {
            while ( ( $key, $value ) = each %found ) {
                write(REPORT);
            }
        }
    }
    if ($save_log) {
        print REPORT "-" x 76, "\n";
        close(REPORT) if ( fileno(REPORT) );
    }

    # reset things
    $count        = 0;
    $num_scanned  = 0;
    $num_so_far   = 0;
    %found        = ();
    %dirs_scanned = ();
    @files        = ();
    @quoted       = ();
}

format REPORT =
@<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<@>>>>>>>>>>>>>>>>>>>>>>>>>>>>
$key,                                                                 $value 
.

sub clear_output {
    return if ( scalar(@files) > 0 );
    @{ $slist->{data} } = ();
    $window->resize( 630, 325 );
    $window->queue_draw;
    $left_status->set_text("Files Scanned: ");
    $mid_status->set_text("Viruses Found: ");
    $right_status->set_text("Ready");
    $top_label->set_text("");
}

sub update {
    if ( $> != 0 ) {
        $top_label->set_text("You must be root to install updates.");
        return;
    }

    $top_label->set_text("Please wait, checking for updates...");
    Gtk2->main_iteration while Gtk2->events_pending;
    $main_vbox->queue_draw;
    $window->queue_draw;
    my @result;

    eval {
        local $SIG{ALRM} = sub { die "failed" };
        alarm 15;

        @result = `$FRESHPATH --stdout`;
        alarm 0;
    };
    if ( $@ && $@ eq "failed" ) {
        $top_label->set_text("Unable to retrieve updates. Try again later.");
        return;
    }
    my $showthis;
    if ( !@result ) {
        $top_label->set_text("Unable to retrieve updates. Try again later.");
    }
    else {
        foreach my $line (@result) {
            if ( $line =~ /Database updated .(\d+) signatures/ ) {
                $showthis
                    = "Your virus signatures have been updated ($1 signatures).";
                $top_label->set_text($showthis);
                last;
            }
        }
        $top_label->set_text("Your virus signatures are up-to-date.")
            if ( !$showthis );
    }
    $window->queue_draw;
}

sub quar_check {
    if ( !-d $v_dir ) {
        $top_label->set_text("No virus directory available.");
        return;
    }
    my @trash = glob "$v_dir/*";
    my $del   = scalar(@trash);
    if ( !$del ) {
        $top_label->set_text("No items currently quarantined.");
    }
    else {
        $top_label->set_text("$del item(s) currently quarantined.");
    }
}

sub del_quar {
    unless ( -e $v_dir ) {
        $top_label->set_text("No quarantine directory to empty.");
        return;
    }
    else {
        my @trash = glob "$v_dir/*";
        if ( scalar(@trash) == 0 ) {
            $top_label->set_text("No items to empty.");
        }
        else {
            my $del = 0;
            foreach (@trash) {
                unlink and $del++;
            }
            $top_label->set_text("Removed $del item(s).");
        }
    }
}

sub move_to_quarantine {
    my $number   = shift;
    my $basename = $virus[$number]{base};
    if ( not -e $v_dir or not -d $v_dir ) {
        $top_label->set_text("Quarantine directory does not exist.");
        return;
    }
    chmod 0600, $virus[$number]{full};
    system( "mv", $virus[$number]{full}, "$v_dir/$basename" );
    rename( "$v_dir/$basename", "$v_dir/$basename.VIRUS" );
    if ( not -e $virus[$number]{full} ) {
        return 1;
    }
    else {
        return -1;
    }
}

#------------------history and history files stuff-------------------
sub history {
    my $choice = shift;
    @h_files = ();
    unless ( -e $l_dir ) {
        $top_label->set_text("History directory not found.");
        return;
    }

    my ( $filename, $dialog );
    if ( $choice eq "view" ) {
        @h_files = glob "$l_dir/*.log";
        if ( scalar(@h_files) == 0 ) {
            $top_label->set_text("No history logs are available to view.");
            return;
        }
        chdir($l_dir);
        $dialog = Gtk2::FileChooserDialog->new(
            'Select a History File', undef, 'open',
            'gtk-cancel' => 'cancel',
            'gtk-ok'     => 'ok'
        );

        if ( "ok" eq $dialog->run ) {
            $filename = $dialog->get_filename;
            $dialog->destroy;
        }
        else {
            $dialog->destroy;
            return;
        }
        my $base = basename($filename);

        $new_win = Gtk2::Dialog->new( "Viewing $base",
            undef, [], 'gtk-close' => 'close' );
        $new_win->set_default_response('close');
        $new_win->signal_connect(
            response => sub { $new_win->destroy; undef $new_win } );
        $new_win->set_default_size( 650, 350 );

        my $textview = Gtk2::TextView->new;
        $textview->set( editable => FALSE );

        open( FILE, "<", $filename )
            or $top_label->set_text("Problems opening $filename...")
            and return;
        my $text;
        {
            local $/ = undef;
            $text = <FILE>;
        }
        close(FILE) or warn "Unable to close FILE $filename! $!\n";

        my $textbuffer = $textview->get_buffer;
        $textbuffer->create_tag( 'mono', family => 'Monospace' );
        $textbuffer->insert_with_tags_by_name( $textbuffer->get_start_iter,
            $text, 'mono' );

        my $scroll_win = Gtk2::ScrolledWindow->new;
        $scroll_win->set_border_width(5);
        $scroll_win->set_shadow_type('etched-in');
        $scroll_win->set_policy( 'automatic', 'automatic' );

        $new_win->vbox->pack_start( $scroll_win, TRUE, TRUE, 0 );

        $scroll_win->add($textview);
        $new_win->show_all();

    }
    elsif ( $choice eq "delete" ) {
        @h_files = glob "$l_dir/*.log";
        if ( scalar(@h_files) == 0 ) {
            $top_label->set_text("No history logs are available to delete.");
            return;
        }
        if ( defined $new_win ) {

            # why this "iconifies" in the first place, I have NO idea...
            # in the meanwhile, we destroy $new_win and then undef it
            $new_win->deiconify;
        }
        else {
            $new_win = Gtk2::Window->new;
            $new_win->signal_connect(
                destroy => sub { $new_win->destroy; undef $new_win } );
            $new_win->set_default_size( 250, 200 );
            $new_win->set_title("Scanning Histories");

            my $new_vbox = Gtk2::VBox->new;
            $new_win->add($new_vbox);

            my $s_win = Gtk2::ScrolledWindow->new;
            $s_win->set_shadow_type('etched-in');
            $s_win->set_policy( 'automatic', 'automatic' );
            $new_vbox->pack_start( $s_win, TRUE, TRUE, 0 );

            $new_hlist = Gtk2::SimpleList->new( 'Histories' => 'text', );
            $s_win->add($new_hlist);

            my $new_hbox = Gtk2::HBox->new;
            $new_vbox->pack_start( $new_hbox, FALSE, FALSE, 0 );

            my $pos_quit = Gtk2::Button->new_with_label("Close Window");
            $new_hbox->pack_start( $pos_quit, TRUE, FALSE, 4 );
            $pos_quit->signal_connect(
                clicked => sub { $new_win->destroy; undef $new_win } );
            my $del_single = Gtk2::Button->new_with_label("Delete");
            $new_hbox->pack_start( $del_single, TRUE, FALSE, 4 );
            $del_single->signal_connect(
                clicked => \&del_single,
                "del_single"
            );
            my $del_all = Gtk2::Button->new_with_label("Delete All");
            $new_hbox->pack_start( $del_all, TRUE, FALSE, 4 );
            $del_all->signal_connect( clicked => \&del_all, "del_all" );

            $h_label = Gtk2::Label->new();
            $new_vbox->pack_start( $h_label, FALSE, FALSE, 2 );

            for my $opt (@h_files) {
                push @{ $new_hlist->{data} }, basename($opt);
            }
            $new_win->show_all;
        }
    }
}

sub del_single {
    my @sel = $new_hlist->get_selected_indices;
    return if ( !@sel );
    my $deref = $sel[0];
    return if ( not exists $h_files[$deref] );
    unlink $h_files[$deref];
    if ( -e $h_files[$deref] ) {
        $q_label->set_text("Unable to delete $h_files[$deref].");
        return;
    }
    splice @{ $new_hlist->{data} }, $deref, 1;
    my $base = basename( $h_files[$deref] );
    $h_label->set_text("Deleted $base.");
    @h_files = glob "$l_dir/*";
}

sub del_all {
    return unless (@h_files);
    my $confirm_message = "Really delete all history logs?";
    my $confirm         = Gtk2::MessageDialog->new_with_markup( $window,
        [qw(modal destroy-with-parent)],
        'question', 'ok-cancel', $confirm_message );

    if ( "cancel" eq $confirm->run ) {
        $confirm->destroy;
        return;
    }
    else {
        $confirm->destroy;
        my @not_del;
        my $size = @h_files;
        foreach (@h_files) {
            unlink($_) or push( @not_del, $_ );
        }
        if ( scalar(@not_del) >= 1 ) {
            $h_label->set_text("Could not delete files: @not_del");
        }
        else {
            $h_label->set_text("Successfully removed history logs.");
        }
        splice @{ $new_hlist->{data} }, 0, $size;
        @h_files = glob "$l_dir/*";
    }
}

#-----------------^history and history files stuff^------------------

sub startup_prefs {

    # this is an effort to combine any
    # and all startup things, such as the upcoming prefs file

    # theoretically, this next line shouldn't be necessary
    if ( $q_state eq "disabled" || $l_state eq "disabled" ) {
        $top_label->set_text(
            "Unable to create ~/.clamtk directory - check permissions.");
    }

    date_diff();
    first_run();    # not needed for debian
}

sub sys_info {
    my ( $info, $return, $version, $number );

    # Signature date info
    if ($DEFPATH) {
        $return = `$SIGPATH --info $DEFPATH 2>&1`;
    }
    else {
        $return = "(No definitions found)";
    }

    if ( $return =~ /^Build time: (\d+\s\w+\s\d{4})/ ) {
        $info = $1;
    }
    else {
        $info = $return;
    }

    # ClamAV version
    my $ver_info = `$CLAMPATH -V`;
    chomp($ver_info);
    ( $version = $ver_info ) =~ s/^(\w+\s+0\.\d+(?:\.\d+)).*$/$1/;

    # Number of signatures
    $number = num_of_sigs();

    my $total = "\nBuild: $version\t\n\n"
        . "Signatures: $number\t\n"
        . "($info)\n\n"
        . "GUI Version: $VERSION\n";
    my $dialog = Gtk2::MessageDialog->new_with_markup( $window,
        [qw(modal destroy-with-parent)],
        'info', "ok", $total );

    if ( "ok" eq $dialog->run ) {
        $dialog->destroy;
        return;
    }
}

sub cvd_check {
    my $return;

    # this path is something to watch
    # (i.e., /var/clamav vs. /var/lib/clamav)
    if ($DEFPATH) {
        $return = `$SIGPATH --info $DEFPATH 2>&1`;
    }
    else {
        return ( 0, 0, 0 );    #no definitions found
    }

    if ( $return =~ /^Build time: (\d+\s\w+\s\d{4})/ ) {
        return $1;
    }
    else {
        $top_label->set_text("Date of ClamAV Signatures: $return");
    }
}

sub date_diff {
    my ( $day1, $month1, $year1 ) = split / /,
        strftime( '%d %m %Y', localtime );
    my ( $day2, $month2, $year2 ) = split / /, cvd_check();
    unless ( $day2 && $month2 && $year2 ) {
        $top_label->set_text("Warning: No virus definitions found!");
        return;
    }
    my %months = (
        'Jan' => 1,
        'Feb' => 2,
        'Mar' => 3,
        'Apr' => 4,
        'May' => 5,
        'Jun' => 6,
        'Jul' => 7,
        'Aug' => 8,
        'Sep' => 9,
        'Oct' => 10,
        'Nov' => 11,
        'Dec' => 12,
    );
    my $diff = Delta_Days( $year1, $month1, $day1, $year2, $months{$month2},
        $day2 );
    if ( $diff <= -5 ) {

        $diff *= -1;    # $diff returns a negative number, so...
        my $warning = "Warning! Signatures are $diff days old!";
        my $dialog  = Gtk2::MessageDialog->new_with_markup( $window,
            [qw(modal destroy-with-parent)],
            'warning', "ok", $warning );

        if ( "ok" eq $dialog->run ) {
            $dialog->destroy;
            return;
        }
    }
    else {
        return;
    }
}

sub first_run {
    return if ( -e "$c_dir/first_run" || $> != 0 );
    my $warning
        = "This is a rebuild for Fedora Extras.\nUnlike Dag's ClamAV rpms,"
        . " the Extras rpms do not\nautomatically edit freshclam.conf and clamd."
        . "conf under /etc.\nPlease edit those before attempting signature updates.\n";
    my $dialog = Gtk2::MessageDialog->new_with_markup( $window,
        [qw(modal destroy-with-parent)],
        'warning', "ok", $warning );
    open( FILE, ">", "$c_dir/first_run" )
        or warn "Couldn't create 'first_run' file...\n";
    close(FILE) or warn "Couldn't close FILE $c_dir first_run! $!\n";

    if ( "ok" eq $dialog->run ) {
        $dialog->destroy;
        return;
    }
}

sub default_check {
    my $action = shift;
    if ( $action eq "ScanHidden" ) {
        $hidden ^= 1;
    }
    if ( $action eq "ShowAll" ) {
        $showall ^= 1;
    }
    if ( $action eq "SaveLog" ) {
        $save_log ^= 1;
    }
    if ( $action eq "Broken" ) {
        $detect_broken ^= 1;
        if ( $detect_broken == 1 ) {
            $command .= " --detect-broken";
        }
        else {
            $command =~ s/\s+--detect-broken//;
        }
    }
}

sub action_actions {
    my ( $previous, $new ) = @_;

    # $previous is the old value; $new is the new value :)
    if ( $new == 1 ) {
        $ren = "None";
        $top_label->set_text("No action will be taken on infected files.");
    }
    elsif ( $new == 2 ) {
        $ren = "Quarantine";
        $top_label->set_text("Infected files will be quarantined.");
    }
    elsif ( $new == 3 ) {
        $ren = "Delete";
        $top_label->set_markup(
            "<b>Warning!</b> Files thought to be infected will be deleted!");
    }
}

sub num_of_sigs {
    my $total;
    if ($SIGPATH) {
        $total = `$SIGPATH -l |wc -l 2>&1`;
        chomp($total);
        return $total;
    }
    else {
        return "Unknown";
    }
}

sub maintenance {
    my $new_win = Gtk2::Window->new;
    $new_win->signal_connect( destroy => sub { $new_win->destroy; } );
    $new_win->set_default_size( 250, 200 );
    $new_win->set_title("Quarantine");

    my $new_vbox = Gtk2::VBox->new( FALSE, 2 );
    $new_win->add($new_vbox);

    @q_files = glob "$v_dir/*";
    my $s_win = Gtk2::ScrolledWindow->new;
    $s_win->set_shadow_type('etched-in');
    $s_win->set_policy( 'automatic', 'automatic' );
    $new_vbox->pack_start( $s_win, TRUE, TRUE, 0 );

    $new_slist = Gtk2::SimpleList->new( 'File' => 'text', );
    $s_win->add($new_slist);

    my $new_hbox = Gtk2::HBox->new( FALSE, 2 );
    $new_vbox->pack_start( $new_hbox, FALSE, FALSE, 0 );

    my $pos_quit = Gtk2::Button->new_with_label("Close Window");
    $new_hbox->pack_start( $pos_quit, TRUE, FALSE, 4 );
    $pos_quit->signal_connect( clicked => sub { $new_win->destroy } );
    my $false_pos = Gtk2::Button->new_with_label("False Positive");
    $new_hbox->pack_start( $false_pos, TRUE, FALSE, 4 );
    $false_pos->signal_connect( clicked => \&false_pos, "false_pos" );
    my $del_pos = Gtk2::Button->new_with_label("Delete");
    $new_hbox->pack_start( $del_pos, TRUE, FALSE, 4 );
    $del_pos->signal_connect( clicked => \&del_pos, "false_pos" );

    $q_label = Gtk2::Label->new();
    $new_vbox->pack_start( $q_label, FALSE, FALSE, 2 );

    for my $opt (@q_files) {
        push @{ $new_slist->{data} }, basename($opt);
    }
    $new_win->show_all;
}

sub false_pos {
    my @sel = $new_slist->get_selected_indices;
    return if ( !@sel );
    my $deref = $sel[0];
    return if ( not exists $q_files[$deref] );
    my $base = basename( $q_files[$deref] );
    system( "mv", $q_files[$deref], $directory );
    my $new_name = $base;
    $new_name =~ s/.VIRUS$//;
    rename( "$directory/$base", "$directory/$new_name" );

    if ( -e $q_files[$deref] ) {
        $q_label->set_text("Unable to move $base.");
        return;
    }
    splice @{ $new_slist->{data} }, $deref, 1;
    $q_label->set_text("$base moved to home directory.");
    @q_files = glob "$v_dir/*";
}

sub del_pos {
    my @sel = $new_slist->get_selected_indices;
    return if ( !@sel );
    my $deref = $sel[0];
    return if ( not exists $q_files[$deref] );
    my $base = basename( $q_files[$deref] );
    unlink $q_files[$deref];
    if ( -e $q_files[$deref] ) {
        $q_label->set_text("Unable to delete $base.");
        return;
    }

    splice @{ $new_slist->{data} }, $deref, 1;
    $q_label->set_text("Deleted $base.");
    @q_files = glob "$v_dir/*";
}

sub confirm {
    my $number         = shift;
    my $do_this        = shift;
    my $current_status = $virus[$number]{status};

    if ( $do_this eq "q" ) {
        if ( not -e $virus[$number]{full} ) {
            $top_label->set_text("File has been moved or deleted already.");
            return;
        }
        if ( move_to_quarantine($number) ) {
            $top_label->set_text(
                "$virus[$number]{base} has been quarantined.");
            $virus[$number]{status} = "$current_status <b>(Quarantined)</b>";
        }
        else {
            $top_label->set_text(
                "$virus[$number]{base} could not be quarantined.");
            return;
        }
    }
    elsif ( $do_this eq "d" ) {
        if ( not -e $virus[$number]{full} ) {
            $top_label->set_text("File has been moved or deleted already.");
            return;
        }
        my $confirm_message = "Really delete $virus[$number]{base}?";
        my $confirm         = Gtk2::MessageDialog->new_with_markup( $window,
            [qw(modal destroy-with-parent)],
            'question', 'ok-cancel', $confirm_message );

        if ( "cancel" eq $confirm->run ) {
            $confirm->destroy;
            return;
        }
        else {
            $confirm->destroy;
            if ( unlink( $virus[$number]{full} ) ) {
                $top_label->set_text("Deleted $virus[$number]{base}.");
                $virus[$number]{status} = "$current_status <b>(Deleted)</b>";
            }
            else {
                $top_label->set_text(
                    "Unable to delete $virus[$number]{base}.");
                return;
            }
        }
    }
    slist_delete($number);
}

sub slist_delete {
    my $number = shift;
    $slist->{data}[$number][0]
        = "<span foreground='#CCCCCC'>$virus[$number]{base}</span>";
    $slist->{data}[$number][1]
        = "<span foreground='#CCCCCC'>$virus[$number]{type}</span>";
    $slist->{data}[$number][2]
        = "<span foreground='#CCCCCC'>$virus[$number]{size}</span>";
    $slist->{data}[$number][3]
        = "<span foreground='#CCCCCC'>$virus[$number]{status}</span>";
    $window->queue_draw;
}

sub type {
    my $file   = shift;
    my $quoted = quotemeta($file);
    my $tmp    = `$FILE -b $quoted`;
    chomp($tmp);
    if ( $tmp =~ /^ERROR/ ) {
        $tmp = `$FILE -b $file`;
        chomp($tmp);
    }

    # once more - file may have been removed already, so...
    if ( $tmp =~ /^ERROR/ ) {
        $tmp = "unknown (e)"    # leave the 'e' so we know there's an error
    }

    # this is an attempt to clean up some file descriptions
    # shortening the length would be easier, but wouldn't look nice
    $tmp =~ s/(.*?),.*$/$1/;    # remove 1st comma and on
    $tmp =~ s/(MS Windows) PE.*?(\w+ executable).*?/$1 $2/i; # these are long!
    $tmp =~ s/(.*?executable).*$/$1/i;        # remove after word 'executable'
    $tmp =~ s/(.*? text).*$/$1/i;             # good for URL descriptors
    $tmp =~ s/Intel 80386\s//g;
    $tmp =~ s/(DBase \d+ data file) .*/$1/;
    $tmp =~ s/extended-ASCII//;

    # sometimes `file` doesn't return anything...
    $tmp ||= "unknown";
    return $tmp;
}

# there are shorter ways to do this, but this is easy to read
sub size {
    my $file = shift;
    my $size = -s $file;
    if ( $size == 0 ) {
        $size = 'empty';
    }
    elsif ( $size < 1024 ) {
        $size = 'less than 1 KB';
    }
    elsif ( $size < 1000 * 1000 ) {
        $size = $size / 1000;
        $size = sprintf "%.2f", ($size);
        $size .= ' KB';
    }
    else {
        my $kb = $size / 1000;
        my $mb = $kb / 1000;
        $size = sprintf "%.2f", ($mb);
        $size .= ' MB';
    }
    return $size;
}

sub create_menu {
    my $tree = [
        _File => {
            item_type => '<Branch>',
            children  => [
                'Scan _File' => {
                    callback    => sub { getfile('file') },
                    accelerator => '<ctrl>F',
                    item_type   => '<StockItem>',
                    extra_data  => 'gtk-new',
                },
                'Scan _Directory' => {
                    callback    => sub { getfile('dir') },
                    accelerator => '<ctrl>D',
                    item_type   => '<StockItem>',
                    extra_data  => 'gtk-directory',
                },
                '_Recursive Scan' => {
                    callback    => sub { getfile('recur') },
                    accelerator => '<ctrl>R',
                    item_type   => '<StockItem>',
                    extra_data  => 'gtk-open',
                },
                Separator => { item_type => '<Separator>', },
                _Exit     => {
                    callback    => sub { Gtk2->main_quit; },
                    accelerator => '<ctrl>X',
                    item_type   => '<StockItem>',
                    extra_data  => 'gtk-quit',
                },
            ],
        },
        _View => {
            item_type => '<Branch>',
            children  => [
                'View Histories' => { callback => sub { history('view') }, },
                'Manage Histories' => {
                    callback   => sub { history('delete') },
                    item_type  => '<StockItem>',
                    extra_data => 'gtk-delete',
                },
                'Clear _Output' => {
                    callback    => \&clear_output,
                    accelerator => '<ctrl>O',
                    item_type   => '<StockItem>',
                    extra_data  => 'gtk-clear',
                },
            ],
        },
        _Options => {
            item_type => '<Branch>',
            children  => [
                'Save to Log' => {
                    callback      => \&default_check,
                    callback_data => 'SaveLog',
                    accelerator   => 'F1',
                    item_type     => '<CheckItem>',
                },
                'Scan Hidden Files (.*)' => {
                    callback      => \&default_check,
                    callback_data => 'ScanHidden',
                    accelerator   => 'F2',
                    item_type     => '<CheckItem>',
                },
                'Show All Files' => {
                    callback      => \&default_check,
                    callback_data => 'ShowAll',
                    accelerator   => 'F3',
                    item_type     => '<CheckItem>',
                },
                'Detect Broken Executables' => {
                    callback      => \&default_check,
                    callback_data => 'Broken',
                    accelerator   => 'F4',
                    item_type     => '<CheckItem>',
                },
            ],
        },
        _Actions => {
            item_type => '<Branch>',
            children  => [
                'No Action (default)' => {
                    callback        => \&action_actions,
                    callback_action => 1,
                    item_type       => '<RadioItem>',
                    groupid         => 1,
                },
                'Quarantine' => {
                    callback        => \&action_actions,
                    callback_action => 2,
                    item_type       => '<RadioItem>',
                    groupid         => 1,
                },
                'Delete (Caution!)' => {
                    callback        => \&action_actions,
                    callback_action => 3,
                    item_type       => '<RadioItem>',
                    groupid         => 1,
                },
            ],
        },
        _Quarantine => {
            item_type => '<Branch>',
            children  => [
                _Status      => { callback => \&quar_check },
                _Maintenance => {
                    callback   => \&maintenance,
                    item_type  => '<StockItem>',
                    extra_data => 'gtk-preferences',
                },
                _Empty => {
                    callback   => \&del_quar,
                    item_type  => '<StockItem>',
                    extra_data => 'gtk-delete',
                }
            ],
        },
        _Help => {
            item_type => '<Branch>',
            children  => [
                '_System Information' => {
                    callback    => \&sys_info,
                    accelerator => '<ctrl>S',
                    item_type   => '<StockItem>',
                    extra_data  => 'gtk-properties',
                },
                '_Update Signatures' => {
                    callback    => \&update,
                    accelerator => '<ctrl>U',
                    item_type   => '<StockItem>',
                    extra_data  => $authenticate,
                },
                _About => {
                    callback    => \&about,
                    accelerator => '<ctrl>A',
                    item_type   => '<StockItem>',
                    extra_data  => 'gtk-about',
                },
            ],
        },
    ];

    return $tree;
}
